@@ -9,7 +9,6 @@ use crate::alloc::alloc::{alloc, dealloc, Layout};
 use crate::alloc::boxed::Box;
 use crate::alloc::{vec, vec::Vec};
 use core::any::{type_name, TypeId};
-use core::cell::UnsafeCell;
 use core::hash::{BuildHasher, BuildHasherDefault, Hasher};
 use core::ops::Deref;
 use core::ptr::{self, NonNull};
@@ -27,12 +26,10 @@ use crate::{align, Access, Component, Query};
 /// [`World`](crate::World).
 pub struct Archetype {
     types: Vec<TypeInfo>,
-    state: TypeIdMap<TypeState>,
+    state: OrderedTypeIdMap<TypeState>,
     len: u32,
     entities: Box<[u32]>,
-    // UnsafeCell allows unique references into `data` to be constructed while shared references
-    // containing the `Archetype` exist
-    data: UnsafeCell<NonNull<u8>>,
+    data: NonNull<u8>,
     data_size: usize,
     /// Maps static bundle types to the archetype that an entity from this archetype is moved to
     /// after removing the components from that bundle.
@@ -62,11 +59,11 @@ impl Archetype {
         let max_align = types.first().map_or(1, |ty| ty.layout.align());
         Self::assert_type_info(&types);
         Self {
-            state: types.iter().map(|ty| (ty.id, TypeState::new(0))).collect(),
+            state: OrderedTypeIdMap::new(types.iter().map(|ty| (ty.id, TypeState::new(0)))),
             types,
             entities: Box::new([]),
             len: 0,
-            data: UnsafeCell::new(NonNull::new(max_align as *mut u8).unwrap()),
+            data: NonNull::new(max_align as *mut u8).unwrap(),
             data_size: 0,
             remove_edges: HashMap::default(),
         }
@@ -97,58 +94,65 @@ impl Archetype {
         self.state.contains_key(&id)
     }
 
-    pub(crate) fn get_base<T: Component>(&self) -> Option<NonNull<T>> {
-        let state = self.state.get(&TypeId::of::<T>())?;
-        Some(unsafe {
-            NonNull::new_unchecked(
-                (*self.data.get()).as_ptr().add(state.offset).cast::<T>() as *mut T
-            )
-        })
+    /// Find the state index associated with `T`, if present
+    pub(crate) fn get_state<T: Component>(&self) -> Option<usize> {
+        self.state.search(&TypeId::of::<T>())
+    }
+
+    /// Get the address of the first `T` component using an index from `get_state::<T>`
+    pub(crate) fn get_base<T: Component>(&self, state: usize) -> NonNull<T> {
+        let (id, state) = self.state.get_from_index(state);
+        assert_eq!(id, &TypeId::of::<T>());
+
+        unsafe {
+            NonNull::new_unchecked(self.data.as_ptr().add(state.offset).cast::<T>() as *mut T)
+        }
     }
 
     /// Get the `T` components of these entities, if present
     ///
     /// Useful for efficient serialization.
     pub fn get<T: Component>(&self) -> Option<ColumnRef<'_, T>> {
-        let ptr = self.get_base::<T>()?;
+        let state = self.get_state::<T>()?;
+        let ptr = self.get_base::<T>(state);
         let column = unsafe { slice::from_raw_parts_mut(ptr.as_ptr(), self.len as usize) };
-        self.borrow::<T>();
+        self.borrow::<T>(state);
         Some(ColumnRef {
             archetype: self,
             column,
         })
     }
 
-    pub(crate) fn borrow<T: Component>(&self) {
-        if self
-            .state
-            .get(&TypeId::of::<T>())
-            .map_or(false, |x| !x.borrow.borrow())
-        {
+    pub(crate) fn borrow<T: Component>(&self, state: usize) {
+        let (id, state) = self.state.get_from_index(state);
+        assert_eq!(id, &TypeId::of::<T>());
+
+        if !state.borrow.borrow() {
             panic!("{} already borrowed uniquely", type_name::<T>());
         }
     }
 
-    pub(crate) fn borrow_mut<T: Component>(&self) {
-        if self
-            .state
-            .get(&TypeId::of::<T>())
-            .map_or(false, |x| !x.borrow.borrow_mut())
-        {
+    pub(crate) fn borrow_mut<T: Component>(&self, state: usize) {
+        let (id, state) = self.state.get_from_index(state);
+        assert_eq!(id, &TypeId::of::<T>());
+
+        if !state.borrow.borrow_mut() {
             panic!("{} already borrowed", type_name::<T>());
         }
     }
 
-    pub(crate) fn release<T: Component>(&self) {
-        if let Some(x) = self.state.get(&TypeId::of::<T>()) {
-            x.borrow.release();
-        }
+    pub(crate) fn release<T: Component>(&self, state: usize) {
+        let (id, state) = self.state.get_from_index(state);
+        assert_eq!(id, &TypeId::of::<T>());
+
+        state.borrow.release();
     }
 
-    pub(crate) fn release_mut<T: Component>(&self) {
-        if let Some(x) = self.state.get(&TypeId::of::<T>()) {
-            x.borrow.release_mut();
-        }
+    pub(crate) fn release_mut<T: Component>(&self, state: usize) {
+        let (id, state) = self.state.get_from_index(state);
+        assert_eq!(id, &TypeId::of::<T>());
+
+        state.borrow.release_mut();
     }
 
     /// Number of entities in this archetype
@@ -208,7 +212,7 @@ impl Archetype {
     ) -> Option<NonNull<u8>> {
         debug_assert!(index <= self.len);
         Some(NonNull::new_unchecked(
-            (*self.data.get())
+            self.data
                 .as_ptr()
                 .add(self.state.get(&ty)?.offset + size * index as usize)
                 .cast::<u8>(),
@@ -218,7 +222,7 @@ impl Archetype {
     /// Every type must be written immediately after this call
     pub(crate) unsafe fn allocate(&mut self, id: u32) -> u32 {
         if self.len as usize == self.entities.len() {
-            self.grow(self.len.max(64));
+            self.grow(64);
         }
 
         self.entities[self.len as usize] = id;
@@ -233,7 +237,8 @@ impl Archetype {
 
     pub(crate) fn reserve(&mut self, additional: u32) {
         if additional > (self.capacity() - self.len()) {
-            self.grow(additional - (self.capacity() - self.len()));
+            let increment = additional - (self.capacity() - self.len());
+            self.grow(increment.max(64));
         }
     }
 
@@ -241,7 +246,14 @@ impl Archetype {
         self.entities.len() as u32
     }
 
-    fn grow(&mut self, increment: u32) {
+    /// Increase capacity by at least `min_increment`
+    fn grow(&mut self, min_increment: u32) {
+        // Double capacity or increase it by `min_increment`, whichever is larger.
+        self.grow_exact(self.capacity().max(min_increment))
+    }
+
+    /// Increase capacity by exactly `increment`
+    fn grow_exact(&mut self, increment: u32) {
         unsafe {
             let old_count = self.len as usize;
             let new_cap = self.entities.len() + increment as usize;
@@ -250,11 +262,11 @@ impl Archetype {
             self.entities = new_entities;
 
             let old_data_size = mem::replace(&mut self.data_size, 0);
-            let mut state =
-                TypeIdMap::with_capacity_and_hasher(self.types.len(), Default::default());
+            let mut new_state =
+                OrderedTypeIdMap::new(self.types.iter().map(|ty| (ty.id, TypeState::new(0))));
             for ty in &self.types {
                 self.data_size = align(self.data_size, ty.layout.align());
-                state.insert(ty.id, TypeState::new(self.data_size));
+                new_state.get_mut(&ty.id).unwrap().offset = self.data_size;
                 self.data_size += ty.layout.size() * new_cap;
             }
             let max_align = self.types.first().map_or(1, |x| x.layout.align());
@@ -269,15 +281,15 @@ impl Archetype {
             if old_data_size != 0 {
                 for ty in &self.types {
                     let old_off = self.state.get(&ty.id).unwrap().offset;
-                    let new_off = state.get(&ty.id).unwrap().offset;
+                    let new_off = new_state.get(&ty.id).unwrap().offset;
                     ptr::copy_nonoverlapping(
-                        (*self.data.get()).as_ptr().add(old_off),
+                        self.data.as_ptr().add(old_off),
                         new_data.as_ptr().add(new_off),
                         ty.layout.size() * old_count,
                     );
                 }
                 dealloc(
-                    (*self.data.get()).as_ptr().cast(),
+                    self.data.as_ptr().cast(),
                     Layout::from_size_align_unchecked(
                         old_data_size,
                         self.types.first().map_or(1, |x| x.layout.align()),
@@ -285,8 +297,8 @@ impl Archetype {
                 );
             }
 
-            self.data = UnsafeCell::new(new_data);
-            self.state = state;
+            self.data = new_data;
+            self.state = new_state;
         }
     }
 
@@ -381,9 +393,10 @@ impl Archetype {
         self.reserve(other.len);
         for info in &self.types {
             let src_off = other.state.get(&info.id()).unwrap().offset;
-            let src = (*other.data.get()).as_ptr().add(src_off);
+            let src = other.data.as_ptr().add(src_off);
             let dst_off = self.state.get(&info.id()).unwrap().offset;
-            let dst = (*self.data.get())
+            let dst = self
+                .data
                 .as_ptr()
                 .add(dst_off + self.len as usize * info.layout.size());
             dst.copy_from_nonoverlapping(src, other.len as usize * info.layout.size())
@@ -409,7 +422,7 @@ impl Drop for Archetype {
         if self.data_size != 0 {
             unsafe {
                 dealloc(
-                    (*self.data.get()).as_ptr().cast(),
+                    self.data.as_ptr().cast(),
                     Layout::from_size_align_unchecked(
                         self.data_size,
                         self.types.first().map_or(1, |x| x.layout.align()),
@@ -464,6 +477,36 @@ impl Hasher for TypeIdHasher {
 /// faster no-op hash.
 pub(crate) type TypeIdMap<V> = HashMap<TypeId, V, BuildHasherDefault<TypeIdHasher>>;
 
+struct OrderedTypeIdMap<V>(Box<[(TypeId, V)]>);
+
+impl<V> OrderedTypeIdMap<V> {
+    fn new(iter: impl Iterator<Item = (TypeId, V)>) -> Self {
+        let mut vals = iter.collect::<Box<[_]>>();
+        vals.sort_unstable_by_key(|(id, _)| *id);
+        Self(vals)
+    }
+
+    fn search(&self, id: &TypeId) -> Option<usize> {
+        self.0.binary_search_by_key(id, |(id, _)| *id).ok()
+    }
+
+    fn contains_key(&self, id: &TypeId) -> bool {
+        self.search(id).is_some()
+    }
+
+    fn get(&self, id: &TypeId) -> Option<&V> {
+        self.search(id).map(move |idx| &self.0[idx].1)
+    }
+
+    fn get_mut(&mut self, id: &TypeId) -> Option<&mut V> {
+        self.search(id).map(move |idx| &mut self.0[idx].1)
+    }
+
+    fn get_from_index(&self, idx: usize) -> &(TypeId, V) {
+        &self.0[idx]
+    }
+}
+
 struct TypeState {
     offset: usize,
     borrow: AtomicBorrow,
@@ -557,13 +600,15 @@ impl<T: Component> Deref for ColumnRef<'_, T> {
 
 impl<T: Component> Drop for ColumnRef<'_, T> {
     fn drop(&mut self) {
-        self.archetype.release::<T>();
+        let state = self.archetype.get_state::<T>().unwrap();
+        self.archetype.release::<T>(state);
     }
 }
 
 impl<T: Component> Clone for ColumnRef<'_, T> {
     fn clone(&self) -> Self {
-        self.archetype.borrow::<T>();
+        let state = self.archetype.get_state::<T>().unwrap();
+        self.archetype.borrow::<T>(state);
         Self {
             archetype: self.archetype,
             column: self.column,
