@@ -7,12 +7,14 @@
 
 use core::any::TypeId;
 use core::marker::PhantomData;
+use core::mem;
 use core::ptr::NonNull;
 use core::slice::Iter as SliceIter;
 
+use crate::alloc::boxed::Box;
 use crate::archetype::Archetype;
 use crate::entities::EntityMeta;
-use crate::{Component, Entity};
+use crate::{Component, Entity, World};
 
 /// A collection of component types to fetch from a [`World`](crate::World)
 pub trait Query {
@@ -30,6 +32,10 @@ pub unsafe trait Fetch<'a>: Sized {
     /// Type of value to be fetched
     type Item;
 
+    /// The type of the data which can be cached to speed up retrieving
+    /// the relevant type states from a matching [`Archetype`]
+    type State: Copy;
+
     /// A value on which `get` may never be called
     fn dangling() -> Self;
 
@@ -37,11 +43,13 @@ pub unsafe trait Fetch<'a>: Sized {
     fn access(archetype: &Archetype) -> Option<Access>;
 
     /// Acquire dynamic borrows from `archetype`
-    fn borrow(archetype: &Archetype);
-    /// Construct a `Fetch` for `archetype` if it should be traversed
-    fn new(archetype: &'a Archetype) -> Option<Self>;
+    fn borrow(archetype: &Archetype, state: Self::State);
+    /// Look up state for `archetype` if it should be traversed
+    fn prepare(archetype: &Archetype) -> Option<Self::State>;
+    /// Construct a `Fetch` for `archetype` based on the associated state
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self;
     /// Release dynamic borrows acquired by `borrow`
-    fn release(archetype: &Archetype);
+    fn release(archetype: &Archetype, state: Self::State);
 
     /// Invoke `f` for every component type that may be borrowed and whether the borrow is unique
     fn for_each_borrow(f: impl FnMut(TypeId, bool));
@@ -77,6 +85,8 @@ pub struct FetchRead<T>(NonNull<T>);
 unsafe impl<'a, T: Component> Fetch<'a> for FetchRead<T> {
     type Item = &'a T;
 
+    type State = usize;
+
     fn dangling() -> Self {
         Self(NonNull::dangling())
     }
@@ -89,14 +99,17 @@ unsafe impl<'a, T: Component> Fetch<'a> for FetchRead<T> {
         }
     }
 
-    fn borrow(archetype: &Archetype) {
-        archetype.borrow::<T>();
+    fn borrow(archetype: &Archetype, state: Self::State) {
+        archetype.borrow::<T>(state);
     }
-    fn new(archetype: &'a Archetype) -> Option<Self> {
-        archetype.get_base::<T>().map(Self)
+    fn prepare(archetype: &Archetype) -> Option<Self::State> {
+        archetype.get_state::<T>()
     }
-    fn release(archetype: &Archetype) {
-        archetype.release::<T>();
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+        Self(archetype.get_base(state))
+    }
+    fn release(archetype: &Archetype, state: Self::State) {
+        archetype.release::<T>(state);
     }
 
     fn for_each_borrow(mut f: impl FnMut(TypeId, bool)) {
@@ -118,6 +131,8 @@ pub struct FetchWrite<T>(NonNull<T>);
 unsafe impl<'a, T: Component> Fetch<'a> for FetchWrite<T> {
     type Item = &'a mut T;
 
+    type State = usize;
+
     fn dangling() -> Self {
         Self(NonNull::dangling())
     }
@@ -130,14 +145,18 @@ unsafe impl<'a, T: Component> Fetch<'a> for FetchWrite<T> {
         }
     }
 
-    fn borrow(archetype: &Archetype) {
-        archetype.borrow_mut::<T>();
+    fn borrow(archetype: &Archetype, state: Self::State) {
+        archetype.borrow_mut::<T>(state);
     }
-    fn new(archetype: &'a Archetype) -> Option<Self> {
-        archetype.get_base::<T>().map(Self)
+    #[allow(clippy::needless_question_mark)]
+    fn prepare(archetype: &Archetype) -> Option<Self::State> {
+        Some(archetype.get_state::<T>()?)
     }
-    fn release(archetype: &Archetype) {
-        archetype.release_mut::<T>();
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+        Self(archetype.get_base::<T>(state))
+    }
+    fn release(archetype: &Archetype, state: Self::State) {
+        archetype.release_mut::<T>(state);
     }
 
     fn for_each_borrow(mut f: impl FnMut(TypeId, bool)) {
@@ -159,6 +178,8 @@ pub struct TryFetch<T>(Option<T>);
 unsafe impl<'a, T: Fetch<'a>> Fetch<'a> for TryFetch<T> {
     type Item = Option<T::Item>;
 
+    type State = Option<T::State>;
+
     fn dangling() -> Self {
         Self(None)
     }
@@ -167,14 +188,21 @@ unsafe impl<'a, T: Fetch<'a>> Fetch<'a> for TryFetch<T> {
         Some(T::access(archetype).unwrap_or(Access::Iterate))
     }
 
-    fn borrow(archetype: &Archetype) {
-        T::borrow(archetype)
+    fn borrow(archetype: &Archetype, state: Self::State) {
+        if let Some(state) = state {
+            T::borrow(archetype, state);
+        }
     }
-    fn new(archetype: &'a Archetype) -> Option<Self> {
-        Some(Self(T::new(archetype)))
+    fn prepare(archetype: &Archetype) -> Option<Self::State> {
+        Some(T::prepare(archetype))
     }
-    fn release(archetype: &Archetype) {
-        T::release(archetype)
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+        Self(state.map(|state| T::execute(archetype, state)))
+    }
+    fn release(archetype: &Archetype, state: Self::State) {
+        if let Some(state) = state {
+            T::release(archetype, state);
+        }
     }
 
     fn for_each_borrow(f: impl FnMut(TypeId, bool)) {
@@ -215,6 +243,8 @@ pub struct FetchWithout<T, F>(F, PhantomData<fn(T)>);
 unsafe impl<'a, T: Component, F: Fetch<'a>> Fetch<'a> for FetchWithout<T, F> {
     type Item = F::Item;
 
+    type State = F::State;
+
     fn dangling() -> Self {
         Self(F::dangling(), PhantomData)
     }
@@ -227,17 +257,20 @@ unsafe impl<'a, T: Component, F: Fetch<'a>> Fetch<'a> for FetchWithout<T, F> {
         }
     }
 
-    fn borrow(archetype: &Archetype) {
-        F::borrow(archetype)
+    fn borrow(archetype: &Archetype, state: Self::State) {
+        F::borrow(archetype, state)
     }
-    fn new(archetype: &'a Archetype) -> Option<Self> {
+    fn prepare(archetype: &Archetype) -> Option<Self::State> {
         if archetype.has::<T>() {
             return None;
         }
-        Some(Self(F::new(archetype)?, PhantomData))
+        F::prepare(archetype)
     }
-    fn release(archetype: &Archetype) {
-        F::release(archetype)
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+        Self(F::execute(archetype, state), PhantomData)
+    }
+    fn release(archetype: &Archetype, state: Self::State) {
+        F::release(archetype, state)
     }
 
     fn for_each_borrow(f: impl FnMut(TypeId, bool)) {
@@ -280,6 +313,8 @@ pub struct FetchWith<T, F>(F, PhantomData<fn(T)>);
 unsafe impl<'a, T: Component, F: Fetch<'a>> Fetch<'a> for FetchWith<T, F> {
     type Item = F::Item;
 
+    type State = F::State;
+
     fn dangling() -> Self {
         Self(F::dangling(), PhantomData)
     }
@@ -292,17 +327,20 @@ unsafe impl<'a, T: Component, F: Fetch<'a>> Fetch<'a> for FetchWith<T, F> {
         }
     }
 
-    fn borrow(archetype: &Archetype) {
-        F::borrow(archetype)
+    fn borrow(archetype: &Archetype, state: Self::State) {
+        F::borrow(archetype, state)
     }
-    fn new(archetype: &'a Archetype) -> Option<Self> {
+    fn prepare(archetype: &Archetype) -> Option<Self::State> {
         if !archetype.has::<T>() {
             return None;
         }
-        Some(Self(F::new(archetype)?, PhantomData))
+        F::prepare(archetype)
     }
-    fn release(archetype: &Archetype) {
-        F::release(archetype)
+    fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+        Self(F::execute(archetype, state), PhantomData)
+    }
+    fn release(archetype: &Archetype, state: Self::State) {
+        F::release(archetype, state)
     }
 
     fn for_each_borrow(f: impl FnMut(TypeId, bool)) {
@@ -354,14 +392,12 @@ impl<'w, Q: Query> QueryBorrow<'w, Q> {
 
     fn borrow(&mut self) {
         if self.borrowed {
-            panic!(
-                "called QueryBorrow::iter twice on the same borrow; construct a new query instead"
-            );
+            return;
         }
         for x in self.archetypes {
             // TODO: Release prior borrows on failure?
-            if Q::Fetch::access(x) >= Some(Access::Read) {
-                Q::Fetch::borrow(x);
+            if let Some(state) = Q::Fetch::prepare(x) {
+                Q::Fetch::borrow(x, state);
             }
         }
         self.borrowed = true;
@@ -436,8 +472,8 @@ impl<'w, Q: Query> Drop for QueryBorrow<'w, Q> {
     fn drop(&mut self) {
         if self.borrowed {
             for x in self.archetypes {
-                if Q::Fetch::access(x) >= Some(Access::Read) {
-                    Q::Fetch::release(x);
+                if let Some(state) = Q::Fetch::prepare(x) {
+                    Q::Fetch::release(x, state);
                 }
             }
         }
@@ -465,7 +501,7 @@ impl<'q, Q: Query> QueryIter<'q, Q> {
     ///
     /// `'q` must be sufficient to guarantee that `Q` cannot violate borrow safety, either with
     /// dynamic borrow checks or by representing exclusive access to the `World`.
-    pub(crate) unsafe fn new(meta: &'q [EntityMeta], archetypes: SliceIter<'q, Archetype>) -> Self {
+    unsafe fn new(meta: &'q [EntityMeta], archetypes: SliceIter<'q, Archetype>) -> Self {
         Self {
             meta,
             archetypes,
@@ -486,13 +522,14 @@ impl<'q, Q: Query> Iterator for QueryIter<'q, Q> {
             match unsafe { self.iter.next() } {
                 None => {
                     let archetype = self.archetypes.next()?;
-                    self.iter =
-                        Q::Fetch::new(archetype).map_or(ChunkIter::empty(), |fetch| ChunkIter {
-                            entities: archetype.entities(),
-                            fetch,
-                            position: 0,
-                            len: archetype.len() as usize,
-                        });
+                    let state = Q::Fetch::prepare(archetype);
+                    let fetch = state.map(|state| Q::Fetch::execute(archetype, state));
+                    self.iter = fetch.map_or(ChunkIter::empty(), |fetch| ChunkIter {
+                        entities: archetype.entities(),
+                        fetch,
+                        position: 0,
+                        len: archetype.len() as usize,
+                    });
                     continue;
                 }
                 Some((id, components)) => {
@@ -532,21 +569,8 @@ pub struct QueryMut<'q, Q: Query> {
 
 impl<'q, Q: Query> QueryMut<'q, Q> {
     pub(crate) fn new(meta: &'q [EntityMeta], archetypes: &'q mut [Archetype]) -> Self {
-        // This looks like an ugly O(n^2) loop, but everything's constant after inlining, so in
-        // practice LLVM optimizes it out entirely.
-        let mut i = 0;
-        Q::Fetch::for_each_borrow(|a, unique| {
-            if unique {
-                let mut j = 0;
-                Q::Fetch::for_each_borrow(|b, _| {
-                    if i != j {
-                        core::assert!(a != b, "query violates a unique borrow");
-                    }
-                    j += 1;
-                })
-            }
-            i += 1;
-        });
+        assert_borrow::<Q>();
+
         Self {
             iter: unsafe { QueryIter::new(meta, archetypes.iter()) },
         }
@@ -584,6 +608,24 @@ impl<'q, Q: Query> IntoIterator for QueryMut<'q, Q> {
     }
 }
 
+fn assert_borrow<Q: Query>() {
+    // This looks like an ugly O(n^2) loop, but everything's constant after inlining, so in
+    // practice LLVM optimizes it out entirely.
+    let mut i = 0;
+    Q::Fetch::for_each_borrow(|a, unique| {
+        if unique {
+            let mut j = 0;
+            Q::Fetch::for_each_borrow(|b, _| {
+                if i != j {
+                    core::assert!(a != b, "query violates a unique borrow");
+                }
+                j += 1;
+            })
+        }
+        i += 1;
+    });
+}
+
 struct ChunkIter<Q: Query> {
     entities: NonNull<u32>,
     fetch: Q::Fetch,
@@ -631,7 +673,7 @@ impl<'q, Q: Query> BatchedIter<'q, Q> {
     ///
     /// `'q` must be sufficient to guarantee that `Q` cannot violate borrow safety, either with
     /// dynamic borrow checks or by representing exclusive access to the `World`.
-    pub(crate) unsafe fn new(
+    unsafe fn new(
         meta: &'q [EntityMeta],
         archetypes: SliceIter<'q, Archetype>,
         batch_size: u32,
@@ -662,7 +704,9 @@ impl<'q, Q: Query> Iterator for BatchedIter<'q, Q> {
                 self.batch = 0;
                 continue;
             }
-            if let Some(fetch) = Q::Fetch::new(archetype) {
+            let state = Q::Fetch::prepare(archetype);
+            let fetch = state.map(|state| Q::Fetch::execute(archetype, state));
+            if let Some(fetch) = fetch {
                 self.batch += 1;
                 return Some(Batch {
                     meta: self.meta,
@@ -714,6 +758,8 @@ macro_rules! tuple_impl {
         unsafe impl<'a, $($name: Fetch<'a>),*> Fetch<'a> for ($($name,)*) {
             type Item = ($($name::Item,)*);
 
+            type State = ($($name::State,)*);
+
             fn dangling() -> Self {
                 ($($name::dangling(),)*)
             }
@@ -727,17 +773,24 @@ macro_rules! tuple_impl {
                 Some(access)
             }
 
-            #[allow(unused_variables)]
-            fn borrow(archetype: &Archetype) {
-                $($name::borrow(archetype);)*
+            #[allow(unused_variables, non_snake_case)]
+            fn borrow(archetype: &Archetype, state: Self::State) {
+                let ($($name,)*) = state;
+                $($name::borrow(archetype, $name);)*
             }
             #[allow(unused_variables)]
-            fn new(archetype: &'a Archetype) -> Option<Self> {
-                Some(($($name::new(archetype)?,)*))
+            fn prepare(archetype: &Archetype) -> Option<Self::State> {
+                Some(($($name::prepare(archetype)?,)*))
             }
-            #[allow(unused_variables)]
-            fn release(archetype: &Archetype) {
-                $($name::release(archetype);)*
+            #[allow(unused_variables, non_snake_case)]
+            fn execute(archetype: &'a Archetype, state: Self::State) -> Self {
+                let ($($name,)*) = state;
+                ($($name::execute(archetype, $name),)*)
+            }
+            #[allow(unused_variables, non_snake_case)]
+            fn release(archetype: &Archetype, state: Self::State) {
+                let ($($name,)*) = state;
+                $($name::release(archetype, $name);)*
             }
 
             #[allow(unused_variables, unused_mut)]
@@ -762,6 +815,196 @@ macro_rules! tuple_impl {
 //smaller_tuples_too!(tuple_impl, B, A);
 smaller_tuples_too!(tuple_impl, O, N, M, L, K, J, I, H, G, F, E, D, C, B, A);
 
+/// A prepared query can be stored independently of the [`World`] to amortize query set-up costs.
+pub struct PreparedQuery<Q: Query> {
+    memo: (u64, u64),
+    state: Box<[(usize, <Q::Fetch as Fetch<'static>>::State)]>,
+}
+
+impl<Q: Query> Default for PreparedQuery<Q> {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
+impl<Q: Query> PreparedQuery<Q> {
+    /// Create a prepared query which is not yet attached to any world
+    pub fn new() -> Self {
+        Self {
+            // This memo will not match any world as the first ID will 1.
+            memo: (0, 0),
+            state: Default::default(),
+        }
+    }
+
+    #[cold]
+    fn prepare(world: &World) -> Self {
+        let memo = world.memo();
+
+        let state = world
+            .archetypes()
+            .enumerate()
+            .filter_map(|(idx, x)| Q::Fetch::prepare(x).map(|state| (idx, state)))
+            .collect();
+
+        Self { memo, state }
+    }
+
+    /// Query `world`, using dynamic borrow checking
+    ///
+    /// This will panic if it would violate an existing unique reference
+    /// or construct an invalid unique reference.
+    pub fn query<'q>(&'q mut self, world: &'q World) -> PreparedQueryBorrow<'q, Q> {
+        if self.memo != world.memo() {
+            *self = Self::prepare(world);
+        }
+
+        let meta = world.entities_meta();
+        let archetypes = world.archetypes_inner();
+
+        PreparedQueryBorrow::new(meta, archetypes, &*self.state)
+    }
+
+    /// Query a uniquely borrowed world
+    ///
+    /// This avoids the cost of the dynamic borrow checking.
+    /// that is required by [`borrow`][Self::borrow]
+    pub fn query_mut<'q>(&'q mut self, world: &'q mut World) -> PreparedQueryIter<'q, Q> {
+        assert_borrow::<Q>();
+
+        if self.memo != world.memo() {
+            *self = Self::prepare(world);
+        }
+
+        let meta = world.entities_meta();
+        let archetypes = world.archetypes_inner();
+
+        let state: &'q [(usize, <Q::Fetch as Fetch<'q>>::State)] =
+            unsafe { mem::transmute(&*self.state) };
+
+        unsafe { PreparedQueryIter::new(meta, archetypes, state.iter()) }
+    }
+}
+
+/// Combined borrow of a [`PreparedQuery`] and a [`World`]
+pub struct PreparedQueryBorrow<'q, Q: Query> {
+    meta: &'q [EntityMeta],
+    archetypes: &'q [Archetype],
+    state: &'q [(usize, <Q::Fetch as Fetch<'static>>::State)],
+}
+
+impl<'q, Q: Query> PreparedQueryBorrow<'q, Q> {
+    fn new(
+        meta: &'q [EntityMeta],
+        archetypes: &'q [Archetype],
+        state: &'q [(usize, <Q::Fetch as Fetch<'static>>::State)],
+    ) -> Self {
+        for (idx, state) in state {
+            Q::Fetch::borrow(&archetypes[*idx], *state);
+        }
+
+        Self {
+            meta,
+            archetypes,
+            state,
+        }
+    }
+
+    /// Execute the prepared query
+    // The lifetime narrowing here is required for soundness.
+    pub fn iter<'i>(&'i mut self) -> PreparedQueryIter<'i, Q> {
+        let state: &'i [(usize, <Q::Fetch as Fetch<'i>>::State)] =
+            unsafe { mem::transmute(self.state) };
+
+        unsafe { PreparedQueryIter::new(self.meta, self.archetypes, state.iter()) }
+    }
+}
+
+impl<Q: Query> Drop for PreparedQueryBorrow<'_, Q> {
+    fn drop(&mut self) {
+        for (idx, state) in self.state {
+            Q::Fetch::release(&self.archetypes[*idx], *state);
+        }
+    }
+}
+
+/// Iterates over all entities matching a [`PreparedQuery`]
+pub struct PreparedQueryIter<'q, Q: Query> {
+    meta: &'q [EntityMeta],
+    archetypes: &'q [Archetype],
+    state: SliceIter<'q, (usize, <Q::Fetch as Fetch<'q>>::State)>,
+    iter: ChunkIter<Q>,
+}
+
+impl<'q, Q: Query> PreparedQueryIter<'q, Q> {
+    /// # Safety
+    ///
+    /// `'q` must be sufficient to guarantee that `Q` cannot violate borrow safety, either with
+    /// dynamic borrow checks or by representing exclusive access to the `World`.
+    unsafe fn new(
+        meta: &'q [EntityMeta],
+        archetypes: &'q [Archetype],
+        state: SliceIter<'q, (usize, <Q::Fetch as Fetch<'q>>::State)>,
+    ) -> Self {
+        Self {
+            meta,
+            archetypes,
+            state,
+            iter: ChunkIter::empty(),
+        }
+    }
+}
+
+unsafe impl<Q: Query> Send for PreparedQueryIter<'_, Q> {}
+unsafe impl<Q: Query> Sync for PreparedQueryIter<'_, Q> {}
+
+impl<'q, Q: Query> Iterator for PreparedQueryIter<'q, Q> {
+    type Item = (Entity, QueryItem<'q, Q>);
+
+    #[inline(always)]
+    fn next(&mut self) -> Option<Self::Item> {
+        loop {
+            match unsafe { self.iter.next() } {
+                None => {
+                    let (idx, state) = self.state.next()?;
+                    let archetype = &self.archetypes[*idx];
+                    self.iter = ChunkIter {
+                        entities: archetype.entities(),
+                        fetch: Q::Fetch::execute(archetype, *state),
+                        position: 0,
+                        len: archetype.len() as usize,
+                    };
+                    continue;
+                }
+                Some((id, components)) => {
+                    return Some((
+                        Entity {
+                            id,
+                            generation: unsafe { self.meta.get_unchecked(id as usize).generation },
+                        },
+                        components,
+                    ));
+                }
+            }
+        }
+    }
+
+    fn size_hint(&self) -> (usize, Option<usize>) {
+        let n = self.len();
+        (n, Some(n))
+    }
+}
+
+impl<Q: Query> ExactSizeIterator for PreparedQueryIter<'_, Q> {
+    fn len(&self) -> usize {
+        self.state
+            .clone()
+            .map(|(idx, _)| self.archetypes[*idx].len() as usize)
+            .sum::<usize>()
+            + self.iter.remaining()
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
