@@ -9,6 +9,7 @@ use crate::alloc::{vec, vec::Vec};
 use core::any::TypeId;
 use core::borrow::Borrow;
 use core::convert::TryFrom;
+use core::sync::atomic::{AtomicU64, Ordering};
 use core::{fmt, mem, ptr};
 
 #[cfg(feature = "std")]
@@ -18,7 +19,7 @@ use hashbrown::{HashMap, HashSet};
 
 use crate::alloc::boxed::Box;
 use crate::archetype::{Archetype, TypeIdMap, TypeInfo};
-use crate::entities::{Entities, Location, ReserveEntitiesIterator};
+use crate::entities::{Entities, EntityMeta, Location, ReserveEntitiesIterator};
 use crate::{
     Bundle, ColumnBatch, DynamicBundle, Entity, EntityRef, Fetch, MissingComponent, NoSuchEntity,
     Query, QueryBorrow, QueryItem, QueryMut, QueryOne, Ref, RefMut,
@@ -47,15 +48,21 @@ pub struct World {
     archetypes: ArchetypeSet,
     /// Maps statically-typed bundle types to archetypes
     bundle_to_archetype: TypeIdMap<u32>,
+    id: u64,
 }
 
 impl World {
     /// Create an empty world
     pub fn new() -> Self {
+        static ID: AtomicU64 = AtomicU64::new(1);
+
         Self {
             entities: Entities::default(),
             archetypes: ArchetypeSet::new(),
             bundle_to_archetype: HashMap::default(),
+            id: ID
+                .fetch_update(Ordering::Relaxed, Ordering::Relaxed, |id| id.checked_add(1))
+                .unwrap(),
         }
     }
 
@@ -152,9 +159,10 @@ impl World {
         }
     }
 
-    /// Efficiently spawn a large number of entities with the same components
+    /// Efficiently spawn a large number of entities with the same statically-typed components
     ///
-    /// Faster than calling [`spawn`](Self::spawn) repeatedly with the same components.
+    /// Faster than calling [`spawn`](Self::spawn) repeatedly with the same components, but requires
+    /// that component types are known at compile time.
     ///
     /// # Example
     /// ```
@@ -191,7 +199,7 @@ impl World {
     /// Super-efficiently spawn the contents of a [`ColumnBatch`]
     ///
     /// The fastest, but most specialized, way to spawn large numbers of entities. Useful for high
-    /// performance deserialization.
+    /// performance deserialization. Supports dynamic component types.
     pub fn spawn_column_batch(&mut self, batch: ColumnBatch) -> SpawnColumnBatchIter<'_> {
         self.flush();
 
@@ -285,7 +293,7 @@ impl World {
         Ok(())
     }
 
-    /// Ensure `additional` entities with exact components `T` can be spawned without reallocating
+    /// Ensure at least `additional` entities with exact components `T` can be spawned without reallocating
     pub fn reserve<T: Bundle + 'static>(&mut self, additional: u32) {
         self.reserve_inner::<T>(additional);
     }
@@ -376,6 +384,18 @@ impl World {
         QueryMut::new(&self.entities.meta, &mut self.archetypes.archetypes)
     }
 
+    pub(crate) fn memo(&self) -> (u64, u64) {
+        (self.id, self.archetypes.generation)
+    }
+
+    pub(crate) fn entities_meta(&self) -> &[EntityMeta] {
+        &self.entities.meta
+    }
+
+    pub(crate) fn archetypes_inner(&self) -> &[Archetype] {
+        &self.archetypes.archetypes
+    }
+
     /// Prepare a query against a single entity, using dynamic borrow checking
     ///
     /// Prefer [`query_one_mut`](Self::query_one_mut) when concurrent access to the [`World`] is not
@@ -419,8 +439,9 @@ impl World {
     ) -> Result<QueryItem<'_, Q>, QueryOneError> {
         let loc = self.entities.get(entity)?;
         unsafe {
-            let fetch = Q::Fetch::new(&self.archetypes.archetypes[loc.archetype as usize])
-                .ok_or(QueryOneError::Unsatisfied)?;
+            let archetype = &self.archetypes.archetypes[loc.archetype as usize];
+            let state = Q::Fetch::prepare(archetype).ok_or(QueryOneError::Unsatisfied)?;
+            let fetch = Q::Fetch::execute(archetype, state);
             Ok(fetch.get(loc.index as usize))
         }
     }
@@ -454,6 +475,7 @@ impl World {
         unsafe {
             Ok(EntityRef::new(
                 &self.archetypes.archetypes[loc.archetype as usize],
+                entity.generation,
                 loc.index,
             ))
         }
@@ -480,7 +502,7 @@ impl World {
     /// let mut world = World::new();
     /// let a = world.spawn(());
     /// let b = world.spawn(());
-    /// let ids = world.iter().map(|(id, _)| id).collect::<Vec<_>>();
+    /// let ids = world.iter().map(|entity_ref| entity_ref.entity()).collect::<Vec<_>>();
     /// assert_eq!(ids.len(), 2);
     /// assert!(ids.contains(&a));
     /// assert!(ids.contains(&b));
@@ -694,9 +716,12 @@ impl World {
         if loc.archetype == 0 {
             return Err(MissingComponent::new::<T>().into());
         }
-        Ok(&*self.archetypes.archetypes[loc.archetype as usize]
-            .get_base::<T>()
-            .ok_or_else(MissingComponent::new::<T>)?
+        let archetype = &self.archetypes.archetypes[loc.archetype as usize];
+        let state = archetype
+            .get_state::<T>()
+            .ok_or_else(MissingComponent::new::<T>)?;
+        Ok(&*archetype
+            .get_base::<T>(state)
             .as_ptr()
             .add(loc.index as usize))
     }
@@ -717,9 +742,12 @@ impl World {
         if loc.archetype == 0 {
             return Err(MissingComponent::new::<T>().into());
         }
-        Ok(&mut *self.archetypes.archetypes[loc.archetype as usize]
-            .get_base::<T>()
-            .ok_or_else(MissingComponent::new::<T>)?
+        let archetype = &self.archetypes.archetypes[loc.archetype as usize];
+        let state = archetype
+            .get_state::<T>()
+            .ok_or_else(MissingComponent::new::<T>)?;
+        Ok(&mut *archetype
+            .get_base::<T>(state)
             .as_ptr()
             .add(loc.index as usize))
     }
@@ -738,7 +766,7 @@ impl World {
     /// Useful for dynamically scheduling concurrent queries by checking borrows in advance, and for
     /// efficient serialization.
     pub fn archetypes(&self) -> impl ExactSizeIterator<Item = &'_ Archetype> + '_ {
-        self.archetypes.archetypes.iter()
+        self.archetypes_inner().iter()
     }
 
     /// Returns a distinct value after `archetypes` is changed
@@ -787,7 +815,7 @@ impl Default for World {
 
 impl<'a> IntoIterator for &'a World {
     type IntoIter = Iter<'a>;
-    type Item = (Entity, EntityRef<'a>);
+    type Item = EntityRef<'a>;
     fn into_iter(self) -> Iter<'a> {
         self.iter()
     }
@@ -893,7 +921,7 @@ unsafe impl Send for Iter<'_> {}
 unsafe impl Sync for Iter<'_> {}
 
 impl<'a> Iterator for Iter<'a> {
-    type Item = (Entity, EntityRef<'a>);
+    type Item = EntityRef<'a>;
     fn next(&mut self) -> Option<Self::Item> {
         loop {
             match self.current {
@@ -909,13 +937,9 @@ impl<'a> Iterator for Iter<'a> {
                     let index = self.index;
                     self.index += 1;
                     let id = current.entity_id(index);
-                    return Some((
-                        Entity {
-                            id,
-                            generation: self.entities.meta[id as usize].generation,
-                        },
-                        unsafe { EntityRef::new(current, index) },
-                    ));
+                    return Some(unsafe {
+                        EntityRef::new(current, self.entities.meta[id as usize].generation, index)
+                    });
                 }
             }
         }
